
# as21  assember for S2    28 March 2020
# version 1.0

# update to s2 version 1 ISA  
# add .symbol section      27 January 2025

# type op ={1,2,3,4,5}, label = 10, sym = 11, reg = 12
symtab = {                  # [type,value]
    'ld': [1,1],
    'st': [1,3],
    'mov': [5,5],
    'jmp': [5,6],
    'jal': [3,7],
    'jt': [3,8],
    'jf': [3,9],
    'add': [2,0],
    'sub': [2,1],
    'mul': [2,2],
    'div': [2,3],
    'and': [2,4],
    'or': [2,5],
    'xor': [2,6],
    'eq': [2,7],
    'ne': [2,8],
    'lt': [2,9],       
    'le': [2,10],
    'gt': [2,11],
    'ge': [2,12],
    'shl': [2,13],
    'shr': [2,14],
    'mod': [2,15],
    'ret': [5,19],
    'trap': [5,20],
    'push': [4,21],
    'pop': [4,22],
    'not': [4,23], 
    'r0' : [12,0], 'r1':[12,1], 'r2':[12,2], 'r3':[12,3],
    'r4' : [12,4], 'r5':[12,5], 'r6':[12,6], 'r7':[12,7],
    'r8' : [12,8], 'r9':[12,9], 'r10':[12,10], 'r11':[12,11],
    'r12' : [12,12], 'r13':[12,13], 'r14':[12,14], 'r15':[12,15],  
    'r16' : [12,16], 'r17':[12,17], 'r18':[12,18], 'r19':[12,19],
    'r20' : [12,20], 'r21':[12,21], 'r22':[12,22], 'r23':[12,23],
    'r24' : [12,24], 'r25':[12,25], 'r26':[12,26], 'r27':[12,27],
    'r28' : [12,28], 'r29':[12,29], 'r30':[12,30], 'r31':[12,31]
    }

# fin, fout      file handler

def isnumeric(w):     # still cannot handle #-1
    flag = True
    for c in w:
        if( not c.isdigit()):
            flag = False
    return flag

def isreg(w):
    if w in symtab:
        x = symtab[w]
        if x[0] == 12:
            return True
        if x[0] == 11 and x[1] in range(0,32):  # defined sym
            return True
    return False

def isimm(w):
    return w[0] == '#'

# read input file and recognise three sections
def pass1():
    mode = '*'
    linenum = 0
    ip = 0
    for line in fin:
        linenum += 1
#        print(line,end="")
        tok = line.strip().split()
        if( len(tok) < 1):
            continue
        cmd = tok[0]
        if( cmd == ";"):         # comment line
            continue
        if( cmd == ".code"):
            mode = 'C'
            continue
        if( cmd == ".symbol"):
            mode = 'S'
            continue
        if( cmd == ".data"):
            mode = 'D'
            continue
        if( cmd[0] == ":" ):  # dolable
            symtab[cmd] = [10,ip]
            continue
        if( cmd == ".end"):
            break
        
        if( mode == "S"):
#            print("symbol:",tok)
            symtab[tok[0]] = [11, int(tok[1])]
        elif( mode == "C"):
            ip += 1

#  w can be defined in symtab
def getV(w):
    if w in symtab:
        x = symtab[w]
        if x[0] == 12 or x[0] == 11:  # reg, sym
            return x[1]
    return 0

def getIm(w):
    if w in symtab:    # symbol
        x = symtab[w]
        return x[1]
    return int(w)

def getAds(w):
    if( isnumeric(w)):
        return int(w)               
    if ':'+w in symtab:        # search symtab for label
        return symtab[':'+w][1]
    if w in symtab:            # search for symbol section
        x = symtab[w]
        if x[0] == 11:
            return x[1]
    return 0

def str2(n):        # convert to string pad with blank
    return str(n) + " "

def outL(op,a1,a2):
    s = str2(op)+str2(a1)+str2(a2)
    fout.write('L '+ s + '\n')
    
def outD(op,a1,a2,d):
    s = str2(op)+str2(a1)+str2(a2)+str2(d)
    fout.write('D '+ s + '\n')
    
def outX(op,a1,a2,a3):
    s = str2(a1)+str2(a2)+str2(a3)+str2(op)
    fout.write('X 31 '+ s + '\n')
    
def dogroup1(op,tok):       # ld/st abs,idx,disp
    a2 = tok[2]
    if( a2[0]  == '+'):
        outX(17,getV(tok[1]),getV(a2[1:]),getV(tok[3]))
    elif( a2[0] == '@'):
        outD(op+1,getV(tok[1]),getV(tok[3]),getIm(a2[1:]))
    else:
        outL(op,getV(tok[1]),getAds(tok[2]))
    
def dogroup2(op,tok):       #  op r r r, op r r i, add/sub...
    if( isreg(tok[3])):
        outX(op,getV(tok[1]),getV(tok[2]),getV(tok[3]))
    elif( isimm(tok[3])):
        outD(op+10,getV(tok[1]),getV(tok[2]),getIm(tok[3][1:]))
 
def dogroup3(op,tok):       # jt/jf/jal r ads
    outL(op,getV(tok[1]),getAds(tok[2]))
    
def dogroup4(op,tok):       # op r1 r2, push/pop/not
    if( isreg(tok[1]) and isreg(tok[2])):
#        if ( op == 23 ): # not is a pseudo, use xor r1 r2 #-1
#            outD(16,getV(tok[1]),getV(tok[2]),-1)
#        else:
            outX(op,getV(tok[1]),getV(tok[2]),0)

def dogroup5(op,tok):       # mov jmp trap ret
    if( op == 5):       # mov
        if( isreg(tok[1]) and isreg(tok[2])):
            outX(16,getV(tok[1]),getV(tok[2]),0)
        elif( isreg(tok[1]) and isimm(tok[2])):
            outL(5,getV(tok[1]),getIm(tok[2][1:]))
    elif( op == 6):     # jmp
        outL(op,0,getAds(tok[1]))
    elif( op == 20):    #  trap
        outX(op,getV(tok[1]),getIm(tok[2][1:]),0)
    elif( op == 19):     #  ret
        outX(op,getV(tok[1]),0,0)
                
def pass2():
    mode = '*'
    linenum = 0
    pc = 0
    for line in fin:
        linenum += 1
#        print(linenum,line,end="")
#        print(pc,line,end="")
        fout2.write(str(pc)+" "+line)   # out listing file
        tok = line.strip().split()
        if( len(tok) < 1):
            continue
        cmd = tok[0]
        if( cmd == ";"):         # comment line
            continue
        if( cmd == ".code"):
            mode = 'C'
            continue
        if( cmd == ".symbol"):
            mode = 'S'
            continue
        if( cmd == ".data"):
            mode = 'D'
            fout.write('e\n')  # end code
            fout.write('a ' + tok[1] + '\n')
            continue
        if( cmd[0] == ":" ):
             # dolabel
            continue
        if( cmd == ".end"):
            break
        
        if( mode == "C"):
#            print("code:",tok)
            op = tok[0]
            optype = symtab[op][0]
            opcode = symtab[op][1]
            if( optype == 1):           # group 1-4
                dogroup1(opcode,tok)
            elif( optype == 2):
                dogroup2(opcode,tok)
            elif( optype == 3):
                dogroup3(opcode,tok)
            elif( optype == 4):
                dogroup4(opcode,tok)
            elif( optype == 5):         # group 5
                dogroup5(opcode,tok)
            pc += 1
        elif( mode == "D"):
#            print("data:",tok)
            for x in tok:
                if( x == ';'):
                    break
                fout.write('w ' + x +'\n')

def dumplabel():
    for x in symtab.values():
        if( x[0] == 10 or x[0] == 11 ):
            print(x)
  
def makename(name):
    i = name.index('.')
    return name[:i]
    
def main():
    pass1()
    fin.seek(0)    # rewind
#  print(symtab)
    print("start pass 2")
    fout.write('a 0\n')
    pass2()
    fout.write('e\n')
    fout.write('e\n')   # make sure there is data section
#    dumplabel()

# ---------- initialize --------

fname = input("input:").strip()
foname = makename(fname) + '.obj'
foname2 = makename(fname) + '.lis'

fin = open(fname,"r",encoding='utf8')
fout = open(foname,'w',encoding='utf8')   # object code file
fout2 = open(foname2,'w',encoding='utf8') # listing file
main()
fout.close()
fout2.close()
fin.close()
